@@ -1,7 +1,7 @@
 
 package Convert::ASN1;
 
-# $Id: //depot/asn/lib/Convert/ASN1/_decode.pm#3 $
+# $Id: //depot/asn/lib/Convert/ASN1/_decode.pm#5 $
 
 BEGIN {
   local $SIG{__DIE__};
@@ -38,6 +38,7 @@ sub _decode {
   my $stash = shift;
   my $pos = shift;
   my $end = shift;
+  my $larr = $_[2] || [];
 
   # we try not to copy the input buffer at any time
   foreach my $buf ($_[0]) {
@@ -52,15 +53,18 @@ sub _decode {
       if (length $op->[cTAG]) {
 
 	TAGLOOP: {
-	  my($tag,$len,$npos) = _decode_tl($buf,$pos,$end)
+	  my($tag,$len,$npos,$indef) = _decode_tl($buf,$pos,$end,$larr)
 	    or do {
 	      next OP if $pos==$end and ($idx >= 0 || defined $op->[cOPT]);
 	      die "decode error";
 	    };
 
 	  if ($tag ne $op->[cTAG]) {
-	    next OP if $idx >= 0 || defined $op->[cOPT];
-	    die "decode error";
+	    if ($idx >= 0 || defined $op->[cOPT]) {
+	      unshift @$larr, $len if $indef;
+	      next OP;
+	    }
+	    die "decode error " . unpack("H*",$tag) ."<=>" . unpack("H*",$op->[cTAG]);
 	  }
 
 	  &{$decode[$op->[cTYPE]]}(
@@ -71,10 +75,10 @@ sub _decode {
 	    # should be getting undef. So if it does not get undef
 	    # it knows it has no variable
 	    (($idx >= 0) ? $arr[$idx++] : defined($var) ? $stash->{$var} : 1),
-	    $buf,$npos,$len
+	    $buf,$npos,$len,$larr
 	  );
 
-	  $pos = $npos+$len;
+	  $pos = $npos+$len+$indef;
 
 	  redo TAGLOOP if $idx >= 0 && $pos < $end;
         }
@@ -85,7 +89,7 @@ sub _decode {
 
 	  ANYLOOP: {
 
-	    my($tag,$len,$npos) = _decode_tl($buf,$pos,$end)
+	    my($tag,$len,$npos,$indef) = _decode_tl($buf,$pos,$end,$larr)
 	      or do {
 		next OP if $pos==$end and ($idx >= 0 || defined $op->[cOPT]);
 		die "decode error";
@@ -96,7 +100,7 @@ sub _decode {
 	    (($idx >= 0) ? $arr[$idx++] : $stash->{$var})
 	      = substr($buf,$pos,$len);
 
-	    $pos += $len;
+	    $pos += $len + $indef;
 
 	    redo ANYLOOP if $idx >= 0 && $pos < $end;
 	  }
@@ -104,13 +108,13 @@ sub _decode {
 	else {
 
 	  CHOICELOOP: {
-	    my($tag,$len,$npos) = _decode_tl($buf,$pos,$end)
+	    my($tag,$len,$npos,$indef) = _decode_tl($buf,$pos,$end,$larr)
 	      or do {
 		next OP if $pos==$end and ($idx >= 0 || defined $op->[cOPT]);
 		die "decode error";
 	      };
-
 	    foreach my $cop (@{$op->[cCHILD]}) {
+
 	      if ($tag eq $cop->[cTAG]) {
 
 		my $nstash = $idx >= 0
@@ -123,10 +127,10 @@ sub _decode {
 		  $cop,
 		  $nstash,
 		  $nstash->{$cop->[cVAR]},
-		  $buf,$npos,$len
+		  $buf,$npos,$len,$larr
 		);
 
-		$pos = $npos+$len;
+		$pos = $npos+$len+$indef;
 
 		redo CHOICELOOP if $idx >= 0 && $pos < $end;
 		next OP;
@@ -253,8 +257,8 @@ sub _dec_real {
 
 
 sub _dec_sequence {
-# 0      1    2       3     4     5     6
-# $optn, $op, $stash, $var, $buf, $pos, $len
+# 0      1    2       3     4     5     6     7
+# $optn, $op, $stash, $var, $buf, $pos, $len, $larr
 
   if (defined( my $ch = $_[1]->[cCHILD])) {
     _decode(
@@ -264,7 +268,8 @@ sub _dec_sequence {
       $_[5], #pos
       $_[5]+$_[6], #end
       $_[4], #buf
-      $_[1]->[cLOOP] && $_[1]->[cVAR] #loop
+      $_[1]->[cLOOP] && $_[1]->[cVAR], #loop
+      $_[7]
     );
   }
   else {
@@ -282,10 +287,19 @@ sub _dec_set {
 }
 
 
+my %_dec_time_opt = ( unixtime => 0, withzone => 1, raw => 2);
+
 sub _dec_time {
 # 0      1    2       3     4     5     6
 # $optn, $op, $stash, $var, $buf, $pos, $len
 
+  my $mode = $_dec_time_opt{$_[0]->{'decode_time'} || ''} || 0;
+
+  if ($mode == 2) {
+    $_[3] = substr($_[4],$_[5],$_[6]);
+    return;
+  }
+
   my @bits = (substr($_[4],$_[5],$_[6])
      =~ /^((?:\d\d)?\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)((?:\.\d{1,3})?)(([-+])(\d\d)(\d\d)|Z)/)
      or die "bad time format";
@@ -306,7 +320,7 @@ sub _dec_time {
     $offset = -$offset if $bits[8] eq '-';
     $time += $offset;
   }
-  $_[3] = $_[0]->{'decode_timezone'} ? [$time,$offset] : $time;
+  $_[3] = $mode ? [$time,$offset] : $time;
 }
 
 
@@ -326,7 +340,8 @@ sub _dec_utf8 {
 
 
 sub _decode_tl {
-  my($pos,$end) = @_[1,2];
+  my($pos,$end,$larr) = @_[1,2,3];
+  my $indef = 0;
 
   my $tag = substr($_[0], $pos++, 1);
 
@@ -345,14 +360,71 @@ sub _decode_tl {
   if($len & 0x80) {
     $len &= 0x7f;
 
-    return if $pos+$len > $end ;
+    if ($len) {
+      return if $pos+$len > $end ;
 
-    ($len,$pos) = (unpack("N", "\0" x (4 - $len) . substr($_[0],$pos,$len)), $pos+$len);
+      ($len,$pos) = (unpack("N", "\0" x (4 - $len) . substr($_[0],$pos,$len)), $pos+$len);
+    }
+    else {
+      unless (@$larr) {
+        _scan_indef($_[0],$pos,$end,$larr) or return;
+      }
+      $indef = 2;
+      $len = shift @$larr;
+    }
   }
 
-  return if $pos+$len > $end;
+  return if $pos+$len+$indef > $end;
 
-  ($tag, $len, $pos);
+  ($tag, $len, $pos, $indef);
+}
+
+sub _scan_indef {
+  my($pos,$end,$larr) = @_[1,2,3];
+  @$larr = ();
+  my @depth = ( $pos );
+
+  while(@depth) {
+    return if $pos+2 > $end;
+
+    if (substr($_[0],$pos,2) eq "\0\0") {
+      my $end = $pos;
+      $pos = shift @depth;
+      unshift @$larr, $end-$pos;
+      $pos += 2;
+      next;
+    }
+
+    my $tag = substr($_[0], $pos++, 1);
+
+    if((ord($tag) & 0x1f) == 0x1f) {
+      my $b;
+      my $n=1;
+      do {
+	$tag .= substr($_[0],$pos++,1);
+	$b = ord substr($tag,-1);
+      } while($b & 0x80);
+    }
+    return if $pos >= $end;
+
+    my $len = ord substr($_[0],$pos++,1);
+
+    if($len & 0x80) {
+      if ($len &= 0x7f) {
+	return if $pos+$len > $end ;
+
+	$pos += $len + unpack("N", "\0" x (4 - $len) . substr($_[0],$pos,$len));
+      }
+      else {
+        unshift @depth, $pos;
+      }
+    }
+    else {
+      $pos += $len;
+    }
+  }
+
+  1;
 }
 
 1;
